@@ -1,5 +1,6 @@
 from cv_utils.core import *
-from cv_utils.drawing import *
+from cv_utils.drawing.matplotlib import *
+from cv_utils.drawing.opencv import *
 from cv_utils.color import *
 from cv_utils.video import *
 from cv_utils.calibration import *
